"""Custom types for the library."""

# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from typing import NewType

# pylint: disable=invalid-name

TeamFoundationId = NewType('TeamFoundationId', str)
